# pylint: disable=redefined-outer-name
import os
import shutil
import subprocess
import pytest
from tests.fixtures.app import Matcher, fail, reformatted, success

IGNORE_FILES = (".ignore", ".gitignore", ".git/info/exclude")


@pytest.fixture
def testdir(testfiles):
    return (testfiles / "directory_with_some_not_formatted_files").resolve()


class Repo(os.PathLike):
    def __init__(self, path):
        self.path = path

    def __fspath__(self):
        return str(self.path)

    def __str__(self):
        return str(self.path)

    @property
    def bad1(self):
        return (self.path / "not_formatted_file1.cmake").resolve()

    @property
    def bad2(self):
        return (self.path / "not_formatted_file2.cmake").resolve()

    @property
    def bad3(self):
        return (self.path / "not_formatted_file3.cmake").resolve()

    def gitignore(self, content):
        with open(self.path / ".gitignore", "w", encoding="utf-8") as f:
            f.write(content)

    def named_ignore_file(self, name, content):
        with open(self.path / name, "w", encoding="utf-8") as f:
            f.write(content)


@pytest.fixture
def not_initialized_repo(testdir):
    return Repo(testdir)


@pytest.fixture
def repo(not_initialized_repo):
    subprocess.check_call(["git", "init", "-q"], cwd=not_initialized_repo)
    return not_initialized_repo


def test_gersemi_git_just_works(app, repo):
    assert app("--check", ".", cwd=repo) == fail(
        stdout="",
        stderr=reformatted(repo.bad1, repo.bad2, repo.bad3),
    )


def test_gitignore_empty(app, repo):
    repo.gitignore("")

    assert app("--check", ".", cwd=repo) == fail(
        stdout="",
        stderr=reformatted(repo.bad1, repo.bad2, repo.bad3),
    )


@pytest.mark.parametrize("ignore_file", IGNORE_FILES)
def test_ignore_file_bad1_and_bad3_ignored(app, repo, ignore_file):
    repo.named_ignore_file(
        ignore_file,
        """
not_formatted_file1*
not_formatted_file3*
        """,
    )

    assert app("--check", ".", cwd=repo) == fail(
        stdout="",
        stderr=reformatted(repo.bad2),
    )


@pytest.mark.parametrize("ignore_file", IGNORE_FILES)
def test_dont_respect_ignore_files(app, repo, ignore_file):
    repo.named_ignore_file(
        ignore_file,
        """
not_formatted_file1*
not_formatted_file3*
        """,
    )

    assert app("--no-respect-ignore-files", "--check", ".", cwd=repo) == fail(
        stdout="",
        stderr=reformatted(repo.bad1, repo.bad2, repo.bad3),
    )


def test_gitignore_all_bad_ignored(app, repo):
    repo.gitignore(
        """
not_formatted_file*
        """
    )

    assert app("--check", ".", cwd=repo) == success(stdout="", stderr="")


def test_formatting_of_explicitly_mentioned_ignored_files_fails(app, repo):
    assert app("--check", repo) == fail(
        stdout="",
        stderr=reformatted(repo.bad1, repo.bad2, repo.bad3),
    )
    assert app("--check", repo.bad1, repo.bad2, repo.bad3) == fail(
        stdout="",
        stderr=reformatted(repo.bad1, repo.bad2, repo.bad3),
    )

    repo.gitignore(
        """
not_formatted_file1*
        """
    )

    assert app("--check", repo) == fail(
        stdout="",
        stderr=reformatted(repo.bad2, repo.bad3),
    )
    assert app("--check", repo.bad1, repo.bad2, repo.bad3) == fail(
        stdout="",
        stderr=reformatted(repo.bad1, repo.bad2, repo.bad3),
    )

    repo.gitignore(
        """
not_formatted_file*
        """
    )
    assert app("--check", repo) == success(stdout="", stderr="")
    assert app("--check", repo.bad1, repo.bad2, repo.bad3) == fail(
        stdout="",
        stderr=reformatted(repo.bad1, repo.bad2, repo.bad3),
    )


def starts_with(s):
    return Matcher(
        matcher=lambda inp: inp.startswith(s),
        describe=f'Starts with: "{s}"',
    )


def test_gersemi_in_non_git_repository(app, not_initialized_repo):
    assert app("--check", ".", cwd=not_initialized_repo) == fail(
        stdout="",
        stderr=reformatted(
            not_initialized_repo.bad1,
            not_initialized_repo.bad2,
            not_initialized_repo.bad3,
        ),
    )

    not_initialized_repo.gitignore(
        """
not_formatted_file*
        """
    )

    assert app("--check", ".", cwd=not_initialized_repo) == success(
        stdout="", stderr=""
    )


def test_whitespace_in_filename(app, repo):
    bad4 = (repo.path / "not formatted file 4.cmake").resolve()
    shutil.copy(repo.bad3, bad4)

    assert app("--check", repo) == fail(
        stdout="",
        stderr=reformatted(repo.bad1, repo.bad2, repo.bad3, bad4),
    )

    repo.gitignore(
        """
not formatted file*
        """
    )

    assert app("--check", repo) == fail(
        stdout="",
        stderr=reformatted(repo.bad1, repo.bad2, repo.bad3),
    )


@pytest.fixture
def second_repo(repo):
    result = repo.path.with_name(repo.path.name + "_2")
    assert str(repo) < str(result)

    shutil.copytree(repo, result)
    return Repo(result)


def test_multiple_repositories_in_single_invocation(app, repo, second_repo, testfiles):
    assert app("--check", repo, second_repo) == fail(
        stdout="",
        stderr=reformatted(
            repo.bad1,
            repo.bad2,
            repo.bad3,
            second_repo.bad1,
            second_repo.bad2,
            second_repo.bad3,
        ),
    )

    repo.gitignore(
        """
not_formatted_file1*
not_formatted_file2*
        """
    )
    second_repo.gitignore(
        """
not_formatted_file3*
not_formatted_file2*
        """
    )

    assert app("--check", repo, second_repo) == fail(
        stdout="",
        stderr=reformatted(
            repo.bad3,
            second_repo.bad1,
        ),
    )

    with open(testfiles / ".ignore", "w", encoding="utf-8") as f:
        f.write(
            """
not_formatted_file*
            """
        )

    assert app("--check", repo, second_repo) == success(stdout="", stderr="")



from lark import Token, Tree
from gersemi.ast_helpers import is_newline
from gersemi.transformer import Transformer_InPlace
from gersemi.types import Nodes


class CleanUpComplexArgument(Transformer_InPlace):
    def complex_argument(self, children):
        return Tree("complex_argument", children[1:-1])


class CleanUpNewlines(Transformer_InPlace):
    def newline_or_gap(self, children):
        return Token("NEWLINE", "".join(children)[:2])

    def arguments(self, children: Nodes) -> Tree:
        return Tree("arguments", [child for child in children if not is_newline(child)])


class ParsingTransformer(CleanUpComplexArgument, CleanUpNewlines):
    pass



from typing import Sequence, Tuple, Union
from lark import Tree
from gersemi.ast_helpers import is_comment, is_keyword
from gersemi.keywords import AnyMatcher
from gersemi.specializations.argument_aware_command_invocation_dumper import (
    ArgumentAwareCommandInvocationDumper,
)


def isolate_two_word_keywords(lhs, rhs, children):
    accumulator = []
    for child in children:
        if accumulator:
            if is_comment(child):
                accumulator.append(child)
            elif is_keyword(rhs, child):
                yield Tree("keyword_argument", [*accumulator, child])
                accumulator = []
            else:
                yield from accumulator
                accumulator = [child]
        else:
            if is_keyword(lhs, child):
                accumulator = [child]
            else:
                yield child

    yield from accumulator


class TwoWordKeywordIsolator(ArgumentAwareCommandInvocationDumper):
    _two_words_keywords: Sequence[Tuple[str, Union[str, AnyMatcher]]] = []

    def _preprocess_arguments(self, arguments):
        preprocessed = arguments.children
        for lhs, rhs in self._two_words_keywords:
            preprocessed = isolate_two_word_keywords(lhs, rhs, preprocessed)
        arguments.children = list(preprocessed)
        return arguments

    def keyword_argument(self, tree):
        return self._format_non_option(tree)



import pytest
from gersemi.exceptions import (
    ParsingError,
    UnbalancedBlock,
    UnbalancedBrackets,
    UnbalancedParentheses,
)
from .tests_generator import generate_input_only_tests


def test_parser(parser, case):
    try:
        parser.parse(case.content)
    except ParsingError:
        pytest.fail("invalid input to parse")
        raise


@pytest.mark.parametrize(
    ("invalid_code", "expected_exception"),
    [
        ("set(FOO BAR", UnbalancedParentheses),
        ("message(FOO BAR (BAZ AND FOO)", UnbalancedParentheses),
        ("set(FOO BAR # )", UnbalancedParentheses),
        ("bar)", UnbalancedParentheses),
        ("bar(FOO BAR (BAZ OR FOO)))", UnbalancedParentheses),
        ("another_command #(", UnbalancedParentheses),
        ("foo_command bar_command", UnbalancedParentheses),
        ("foo([[foo]=])", UnbalancedBrackets),
        ("foo([=[bar]])", UnbalancedBrackets),
        ("foo(arg1 arg2 [==[arg3]===] arg4)", UnbalancedBrackets),
        ("macro(foobar)", UnbalancedBlock),
        ("function(foobar)", UnbalancedBlock),
        (
            """function()
            set(FOO foo)""",
            UnbalancedBlock,
        ),
        (
            """function(foobar)
            set(FOO foo)""",
            UnbalancedBlock,
        ),
        (
            """function()
    set(SETTING "FOO")""",
            UnbalancedBlock,
        ),
        (
            """if()
    set(SETTING "FOO")
elseif()
    set(SETTING "BAR")""",
            UnbalancedBlock,
        ),
        (
            """if()
    set(SETTING "FOO")
elseif()
    set(SETTING "BAR")
else()
    set(SETTING "BAR")""",
            UnbalancedBlock,
        ),
        (
            """if()
    set(SETTING "FOO")
else()
    set(SETTING "BAR")""",
            UnbalancedBlock,
        ),
    ],
)
def test_invalid_code_parsing_error(parser, invalid_code, expected_exception):
    with pytest.raises(ParsingError) as exc_info:
        parser.parse(invalid_code)

    assert exc_info.type is expected_exception


pytest_generate_tests = generate_input_only_tests(
    where="parser", input_extension=".cmake"
)



import re
import pytest
from gersemi.builtin_commands import builtin_commands
from gersemi.extension_type import ModuleExtension
from gersemi.extensions import VerificationFailure, verify

defs = "gersemi_foo:command_definitions"
ab = f"{defs}['ab']"
ab_sections = f"{ab}['sections']"
ab_signatures = f"{ab}['signatures']"


@pytest.mark.parametrize(
    "definition",
    [
        {},
        {"ab": {}},
        {"ab": {"options": set()}},
        {"ab": {"options": ()}},
        {"ab": {"options": []}},
        {"ab": {"options": ("ONE", "TWO", "THREE")}},
        {"ab": {"unsupported_entry": {}, "foobar": {}}},
        {"ab": {"unsupported_entry": {}, "options": [], "foobar": {}}},
        {"ab": {"signatures": {"OK": {}}}},
        {"ab": {"signatures": {"OK": {"multi_value_keywords": "VALUES"}}}},
        {"ab": {"options": ("ONE", "22", "THREE")}},
        {"ab": {"one_value_keywords": ("ONE", "234", "THREE")}},
    ],
)
def test_extension_passes_verification(definition):
    try:
        verify(ModuleExtension("foo"), definition)
    except VerificationFailure as e:
        pytest.fail(f"{repr(definition)} should pass verification but doesn't: {e}")


@pytest.mark.parametrize(
    ("definition", "outcome"),
    [
        ((), f"{defs}: is not a mapping"),
        ({12: 34}, f"{defs}: command name (12) has to be a string"),
        (
            {"": 34},
            f"{defs}: command name ('') has to start with a letter or underscore",
        ),
        (
            {"12": 34},
            f"{defs}: command name ('12') has to start with a letter or underscore",
        ),
        ({"ab": 34}, f"{defs}['ab']: is not a mapping"),
        (
            {"ab": {"options": {1: 2}}},
            f"{ab}['options']: keyword (1) has to be a string",
        ),
        (
            {"ab": {"options": ("ONE", 2, "THREE")}},
            f"{ab}['options']: keyword (2) has to be a string",
        ),
        (
            {"ab": {"front_positional_arguments": 34}},
            f"{ab}['front_positional_arguments']: is not a collection of strings (34)",
        ),
        (
            {"ab": {"front_positional_arguments": [34]}},
            f"{ab}['front_positional_arguments'][0]: argument (34) has to be a string",
        ),
        (
            {"ab": {"front_positional_arguments": ["abc", "", 34]}},
            f"{ab}['front_positional_arguments'][2]: argument (34) has to be a string",
        ),
        (
            {"ab": {"sections": ()}},
            f"{ab_sections}: is not a mapping",
        ),
        (
            {"ab": {"sections": {12: 34}}},
            f"{ab_sections}: keyword (12) has to be a string",
        ),
        (
            {"ab": {"sections": {"12": 34}}},
            f"{ab_sections}['12']: is not a mapping",
        ),
        (
            {"ab": {"sections": {"ABC": 34}}},
            f"{ab_sections}['ABC']: is not a mapping",
        ),
        (
            {"ab": {"sections": {"ABC": {"sections": ()}}}},
            f"{ab_sections}['ABC']['sections']: is not a mapping",
        ),
        ({"ab": {"signatures": ()}}, f"{ab_signatures}: is not a mapping"),
        (
            {"ab": {"signatures": {12: 34}}},
            f"{ab_signatures}: signature (12) has to be a string",
        ),
        (
            {"ab": {"signatures": {"12": 34}}},
            f"{ab_signatures}['12']: is not a mapping",
        ),
        (
            {"ab": {"signatures": {"CD": 34}}},
            f"{ab_signatures}['CD']: is not a mapping",
        ),
    ],
)
def test_extension_fails_verification(definition, outcome):
    with pytest.raises(VerificationFailure, match=re.escape(outcome)):
        verify(ModuleExtension("foo"), definition)


@pytest.mark.parametrize(
    ("command", "definition"),
    list(builtin_commands.items()),
    ids=list(builtin_commands),
)
def test_extension_based_on_builtins_works(command, definition):
    try:
        verify(ModuleExtension("foo"), {command: definition})
    except VerificationFailure as e:
        pytest.fail(f"{repr(definition)} should pass verification but doesn't: {e}")



from gersemi.ast_helpers import is_comment, pair
from gersemi.base_dumper import BaseDumper
from gersemi.utils import pop_all


class KeywordWithPairsFormatter(BaseDumper):
    def _pair_arguments(self, args):
        result = []
        accumulator = []
        for arg in args:
            if accumulator:
                accumulator.append(arg)
                if not is_comment(arg):
                    result.append(pair(pop_all(accumulator)))
            else:
                if is_comment(arg):
                    result.append(arg)
                else:
                    accumulator.append(arg)

        if accumulator:
            result.append(pair(pop_all(accumulator)))

        return result

    def _format_keyword_with_pairs(self, args):
        return "\n".join(map(self.visit, self._pair_arguments(args)))



import pytest
from gersemi.configuration import OutcomeConfiguration
from gersemi.custom_command_definition_finder import (
    find_custom_command_definitions,
    get_just_definitions,
)
from gersemi.dumper import Dumper


def create_dumper(definitions):
    return Dumper(
        configuration=OutcomeConfiguration(line_length=80, indent=4),
        known_definitions=definitions,
    )


def test_custom_command_without_keyworded_arguments_formatting(
    parser_with_postprocessing,
    parser,
):
    given = """function(some_custom_command_without_keyworded_arguments only positional arguments)
    message(STATUS "some_custom_command_without_keyworded_arguments")
endfunction()

some_custom_command_without_keyworded_arguments(short positional arguments)

some_custom_command_without_keyworded_arguments(long__________________________________________________ positional__________________________________________________ arguments__________________________________________________)
"""

    expected = """function(
    some_custom_command_without_keyworded_arguments
    only
    positional
    arguments
)
    message(STATUS "some_custom_command_without_keyworded_arguments")
endfunction()

some_custom_command_without_keyworded_arguments(short positional arguments)

some_custom_command_without_keyworded_arguments(
    long__________________________________________________
    positional__________________________________________________
    arguments__________________________________________________
)
"""

    parsed = parser.parse(given)
    definitions = get_just_definitions(find_custom_command_definitions(parsed))
    dumper = create_dumper(definitions)

    parsed_again = parser_with_postprocessing.parse(given, definitions)
    formatted = dumper.visit(parsed_again)

    assert formatted == expected


def test_custom_command_without_keyworded_arguments_with_disabled_reformatting(
    parser_with_postprocessing,
    parser,
):
    given = """function(some_custom_command_without_keyworded_arguments only positional arguments)
    # gersemi: ignore
    message(STATUS "some_custom_command_without_keyworded_arguments")
endfunction()

some_custom_command_without_keyworded_arguments(short positional arguments)

some_custom_command_without_keyworded_arguments(long__________________________________________________ positional__________________________________________________ arguments__________________________________________________)
"""

    expected = """function(
    some_custom_command_without_keyworded_arguments
    only
    positional
    arguments
)
    # gersemi: ignore
    message(STATUS "some_custom_command_without_keyworded_arguments")
endfunction()

some_custom_command_without_keyworded_arguments(short positional arguments)

some_custom_command_without_keyworded_arguments(long__________________________________________________ positional__________________________________________________ arguments__________________________________________________)
"""

    parsed = parser.parse(given)
    definitions = get_just_definitions(find_custom_command_definitions(parsed))
    dumper = create_dumper(definitions)

    parsed_again = parser_with_postprocessing.parse(given, definitions)
    formatted = dumper.visit(parsed_again)

    assert formatted == expected


@pytest.mark.parametrize(
    "given",
    [
        """function(some_custom_command_without_keyworded_arguments only positional arguments)
endfunction()
""",
        """function(some_custom_command_without_keyworded_arguments only positional arguments)

endfunction()
""",
    ],
)
def test_can_deal_with_empty_body_in_custom_command_definition(
    parser_with_postprocessing, parser, given
):
    expected = """function(
    some_custom_command_without_keyworded_arguments
    only
    positional
    arguments
)
endfunction()
"""

    parsed = parser.parse(given)
    definitions = get_just_definitions(find_custom_command_definitions(parsed))
    dumper = create_dumper(definitions)

    parsed_again = parser_with_postprocessing.parse(given, definitions)
    formatted = dumper.visit(parsed_again)

    assert formatted == expected



# pylint: disable=redefined-outer-name
import pytest
from tests.fixtures.app import fail, success
from tests.tests_generator import generate_input_output_tests

LR = "--line-ranges"


@pytest.fixture
def testfiles(testfiles):
    return testfiles / "line_ranges"


def test_line_range_with_no_files_succeeds(app):
    assert app(LR, "123-456") == success()


def test_line_range_with_two_files_fails(app, testfiles):
    given1 = testfiles / "simple--12-20.in.cmake"
    given2 = testfiles / "single-line--7-7.in.cmake"
    assert app(LR, "123-456", "--", given1, given2) == fail(
        stderr="""Line range formatting available only with one source file
""",
    )


def test_line_range_with_directory_with_more_than_one_file_fails(app, testfiles):
    assert app(LR, "123-456", "--", testfiles) == fail(
        stderr="""Line range formatting available only with one source file
""",
    )


def test_multiple_line_range_arguments(app, testfiles):
    given = testfiles / "simple--1-10,22-31.in.cmake"
    with open(testfiles / "simple--1-10,22-31.out.cmake", "r", encoding="utf-8") as f:
        expected = f.read()
        assert app(LR, "1-10,22-31", given) == success(stdout=expected)
        assert app(LR, "1-10", LR, "22-31", given) == success(stdout=expected)
        assert app(LR, "22-31", LR, "1-10", given) == success(stdout=expected)
        assert app(LR, "1-10,22-31", LR, "22-31", given) == success(stdout=expected)
        assert app(LR, "1-10,22-31", LR, "1-10,22-31", given) == success(
            stdout=expected
        )


def test_format_only_line_ranges(app, case):
    _, given_range = case.name.split("--")
    assert app(LR, given_range, "-", input=case.given) == success(stdout=case.expected)


pytest_generate_tests = generate_input_output_tests(
    where="executable/line_ranges",
    input_extension=".in.cmake",
    output_extension=".out.cmake",
)



from dataclasses import astuple, dataclass
from functools import lru_cache
from pathlib import Path
from typing import Sequence
from lark import Lark, UnexpectedInput
from gersemi.exceptions import (
    GenericParsingError,
    UnbalancedBlock,
    UnbalancedBrackets,
    UnbalancedParentheses,
    UnbalancedQuotes,
)
from gersemi.parsing_transformer import ParsingTransformer
from gersemi.postprocessor import postprocess


@dataclass
class CustomBlock:
    start: str
    end: str

    def __hash__(self):
        return hash(astuple(self))

    def rule(self, index) -> str:
        return f"""
        _custom_block__{index}: _block_template{{CB_BEGIN_{index}, CB_END_{index}}}
        CB_BEGIN_{index}: "{self.start.lower()}"i
        CB_END_{index}: "{self.end.lower()}"i

        %extend block: _custom_block__{index}
        """


def create_custom_rules(custom_blocks: Sequence[CustomBlock]) -> str:
    return "\n".join(block.rule(index) for index, block in enumerate(custom_blocks))


@lru_cache(maxsize=None)
def get_lark_parser(grammar_filename, custom_blocks: Sequence[CustomBlock]) -> Lark:
    with open(grammar_filename, encoding="utf-8") as f:
        grammar_as_string = f.read()
        grammar_as_string += create_custom_rules(custom_blocks)

        return Lark(
            grammar=grammar_as_string,
            parser="lalr",
            propagate_positions=False,
            maybe_placeholders=False,
            transformer=ParsingTransformer(),
        )


@lru_cache(maxsize=None)
def block_examples(block_start):
    return (
        f"{block_start}()",
        f"""{block_start}()
foobar()""",
        f"""{block_start}()
foobar(FOO foo)""",
        f"""{block_start}()
set(FOO "foo")""",
        f"""{block_start}()
""",
        f"""{block_start}()
foobar()
""",
        f"""{block_start}()
foobar(FOO foo)
""",
        f"""{block_start}()
set(FOO "foo")
""",
    )


class Parser:
    static_examples = {
        UnbalancedBrackets: [
            "foo(foo [[foo]=]",
            "foo([=[foo bar]])",
            "foo([=[foo bar ]==])",
            "foo(foo [=[foo bar ]==] foo)",
            "foo(foo foo [==[foo]===] foo)",
        ],
        UnbalancedParentheses: [
            "foo(foo foo",
            "foo(foo foo # )",
            "foo(bar",
            "foo(bar\n",
            "foo(BAR (BAZ)",
            "foo(# )",
            "foo(bar))",
            "foo)",
            "foo(BAR (BAZ)))",
            "foo(BAR (BAZ FOO)))",
            "foo",
            "foo # (",
            "foo(commented_argument #foobar)",
            "foo(commented_argument #[[foobar]]",
            "foo(almost_commented_argument #)",
        ],
        UnbalancedBlock: [
            """if()
        else()""",
            """if()
        foo()
        else()"""
            """if()
                    elseif()
            set(FOO foo)""",
            """if()
            foo()
                    elseif()
            set(FOO foo)""",
            """if()
                    elseif()
            set(FOO foo)
                    else()
            set(FOO foo)""",
            """if()
        else()
""",
            """if()
        foo()
        else()
"""
            """if()
                    elseif()
            set(FOO foo)
""",
            """if()
            foo()
                    elseif()
            set(FOO foo)
""",
            """if()
                    elseif()
            set(FOO foo)
                    else()
            set(FOO foo)
""",
        ],
        UnbalancedQuotes: [
            'set(foo ")',
            """set(foo ")
""",
        ],
    }
    static_block_starts = (
        "block",
        "foreach",
        "function",
        "if",
        "macro",
        "while",
    )

    def __init__(self, grammar_filename):
        self.grammar_filename = grammar_filename

    def _match_parsing_error(self, lark_parser, code, custom_blocks, exception):
        specific_error = exception.match_examples(
            lark_parser.parse,
            self.examples(custom_blocks),
            use_accepts=False,
        )

        if not specific_error:
            raise GenericParsingError(
                exception.get_context(code), exception.line, exception.column
            )
        raise specific_error(
            exception.get_context(code), exception.line, exception.column
        )

    def examples(self, custom_blocks):
        result = self.static_examples.copy()
        ub = UnbalancedBlock

        for block_start in self.static_block_starts:
            if ub not in result:
                result[ub] = block_examples(block_start)
            else:
                result[ub].extend(block_examples(block_start))

        for block in custom_blocks:
            if ub not in result:
                result[ub] = block_examples(block.start)
            else:
                result[ub].extend(block_examples(block.start))

        return result

    def parse(  # pylint: disable=inconsistent-return-statements
        self, code, known_definitions=None
    ):
        if known_definitions:
            custom_blocks = tuple(
                CustomBlock(start=n, end=d["block_end"])
                for n, d in known_definitions.items()
                if d.get("block_end", None)
            )
        else:
            custom_blocks = ()

        lark_parser = get_lark_parser(self.grammar_filename, custom_blocks)

        try:
            return lark_parser.parse(code)
        except UnexpectedInput as u:
            self._match_parsing_error(lark_parser, code, custom_blocks, u)


class ParserWithPostProcessing:
    def __init__(self, parser):
        self.parser = parser

    def parse(self, code, known_definitions=None):
        return postprocess(code, known_definitions, self.parser.parse(code))


HERE = Path(__file__).resolve().parent
GRAMMAR = HERE / "cmake.lark"


def create_parser(grammar_filename=GRAMMAR):
    return Parser(grammar_filename)


def create_parser_with_postprocessing(bare_parser):
    return ParserWithPostProcessing(bare_parser)


BARE_PARSER = create_parser()
PARSER = create_parser_with_postprocessing(BARE_PARSER)



from typing import List
from lark import Tree
from gersemi.ast_helpers import (
    is_comment,
    is_line_comment_in,
    is_one_of_keywords,
)
from gersemi.base_command_invocation_dumper import BaseCommandInvocationDumper
from gersemi.configuration import Spaces
from gersemi.types import Nodes
from gersemi.utils import advance


def isolate_unary_operators(operators: List[str], children):
    one_behind = advance(children, times=1, default=None)
    if one_behind is None:
        return

    for current in children:
        if is_one_of_keywords(operators, one_behind):
            if is_comment(current):
                yield Tree("unary_operation", [one_behind])
                yield current
            else:
                yield Tree("unary_operation", [one_behind, current])
            current = advance(children, times=1, default=None)
            if current is None:
                break
        else:
            yield one_behind

        one_behind = current
    else:
        yield one_behind


unary_operators: List[str] = [
    "COMMAND",
    "POLICY",
    "TARGET",
    "TEST",
    "EXISTS",
    "IS_DIRECTORY",
    "IS_SYMLINK",
    "IS_ABSOLUTE",
    "DEFINED",
    "IS_READABLE",
    "IS_WRITABLE",
    "IS_EXECUTABLE",
]
binary_operators: List[str] = [
    "IS_NEWER_THAN",
    "MATCHES",
    "LESS",
    "GREATER",
    "EQUAL",
    "LESS_EQUAL",
    "GREATER_EQUAL",
    "STRLESS",
    "STRGREATER",
    "STREQUAL",
    "STRLESS_EQUAL",
    "STRGREATER_EQUAL",
    "VERSION_LESS",
    "VERSION_GREATER",
    "VERSION_EQUAL",
    "VERSION_LESS_EQUAL",
    "VERSION_GREATER_EQUAL",
    "IN_LIST",
    "PATH_EQUAL",
]


def isolate_binary_tests(children):
    two_behind = advance(children, times=1, default=None)
    if two_behind is None:
        return

    one_behind = advance(children, times=1, default=None)
    if one_behind is None:
        yield two_behind
        return

    for current in children:
        if is_one_of_keywords(binary_operators, one_behind):
            yield Tree("binary_operation", [two_behind, one_behind, current])

            one_behind = advance(children, times=1, default=None)
            if one_behind is None:
                break

            current = advance(children, times=1, default=None)
            if current is None:
                yield one_behind
                break
        else:
            yield two_behind

        two_behind, one_behind = one_behind, current
    else:
        yield two_behind
        yield one_behind


def isolate_conditions(arguments: Nodes) -> Nodes:
    return isolate_unary_operators(
        ["OR"],
        isolate_unary_operators(
            ["AND"],
            isolate_unary_operators(
                ["NOT"],
                isolate_binary_tests(
                    isolate_unary_operators(unary_operators, iter(arguments))
                ),
            ),
        ),
    )


class ConditionSyntaxCommandInvocationDumper(BaseCommandInvocationDumper):
    _inhibit_favour_expansion = True

    def unary_operation(self, tree):
        result = self._try_to_format_into_single_line(tree.children)
        if result is not None:
            return result

        operation, *rest = tree.children
        formatted_operation = self.visit(operation)
        if len(rest) == 0:
            return formatted_operation

        arg, *_ = rest
        if (
            (not is_line_comment_in(operation))
            and isinstance(self.indent_type, Spaces)
            and (len(formatted_operation.strip()) < self.indent_type)
        ):
            return f"{formatted_operation} {self.visit(arg).lstrip()}"

        with self.indented():
            formatted_arg = self.visit(arg)
        return f"{formatted_operation}\n{formatted_arg}"

    def binary_operation(self, tree):
        result = self._try_to_format_into_single_line(tree.children)
        if result is not None:
            return result

        lhs, operation, rhs = tree.children
        formatted_lhs = self.visit(lhs)
        with self.indented():
            formatted_operation = self.visit(operation)
            formatted_rhs = self.visit(rhs)
        return f"{formatted_lhs}\n{formatted_operation}\n{formatted_rhs}"

    def _preprocess_arguments(self, arguments):
        return Tree("arguments", list(isolate_conditions(arguments.children)))

    def _split_arguments(self, arguments):
        if len(arguments) < 1:
            return arguments

        head, *tail = arguments
        return [[head], tail]

    def complex_argument(self, tree):
        arguments, *_ = tree.children
        arguments = self._preprocess_arguments(arguments)
        result = self._try_to_format_into_single_line(
            arguments.children, prefix="(", postfix=")"
        )
        if result is not None:
            return result

        begin = self._indent("(\n")
        with self.indented():
            formatted_arguments = self.visit(arguments)
        end = self._indent(")")
        return f"{begin}{formatted_arguments}\n{end}"


condition_syntax_commands = {
    key: {"__impl": ConditionSyntaxCommandInvocationDumper}
    for key in (
        "elseif",
        "else",
        "endif",
        "endwhile",
        "if",
        "while",
    )
}
